from typing import Optional, Tuple
from html import unescape
import warnings

from anicli_ru._http import client
from anicli_ru.defaults import AniboomDefaults, AniboomM3U8Data


class Aniboom:
    def __init__(self):
        self.session = client
        self.headers = self.session.headers.get("user-agent")

    def get_video_url(self, player_url: str, *, quality: int = 1080, referer: str) -> str:
        """

        :param quality:
        :param player_url:
        :param referer:
        :return:
        """
        warnings.warn("Usage Aniboom.parse() or Aniboom()() methods", category=DeprecationWarning, stacklevel=2)
        return self.parse(player_url, quality=quality, referer=referer)

    @staticmethod
    def get_aniboom_url(raw_aniboom_response: str, *, quality: int = 1080, mpd: bool = False) -> str:
        """
        :param quality: video quality. Available values: 480, 720, 1080
        :param raw_aniboom_response:
        :param mpd: return mpd url extension. Default False
        :return: video url
        """
        warnings.warn("Usage Aniboom.parse() or Aniboom()() methods", category=DeprecationWarning, stacklevel=2)

        return Aniboom._parse_aniboom_response(raw_aniboom_response, quality=quality, mpd=mpd)

    @staticmethod
    def _parse_m3u8(m3u8_url: str) -> Tuple[AniboomM3U8Data, ...]:
        m3u8_response = client.get(m3u8_url, headers={
            "Referer": AniboomDefaults.REFERER, "Accept-Language": AniboomDefaults.ACCEPT_LANG}).text

        return tuple(AniboomM3U8Data(qual, url) for qual, url in AniboomDefaults.RE_M3U8_DATA.findall(m3u8_response))

    @classmethod
    def _set_quality(cls, m3u8_url: str, quality: int = 1080) -> str:
        """set video quality. Works only with m3u8 format

        :param m3u8_url: m3u8 url format
        :param quality: video quality. Default 1080
        :return: video url with set quality
        """
        # parse master.m3u8 response
        results = cls._parse_m3u8(m3u8_url)
        return next((
            m3u8_url.replace("master.m3u8", data.url_suffix) for data in results if data.quality.endswith(str(quality))
        ), m3u8_url)

    @staticmethod
    def is_aniboom(url: str) -> bool:
        """return True if player url is aniboom"""
        return "aniboom" in url

    def _get_aniboom_html_response(self, player_url: str, referer: str) -> str:
        # need set lowercase keys
        r = self.session.get(player_url, headers={"referer": referer,
                                                  "user-agent": self.session.headers["user-agent"]})
        return unescape(r.text)

    @classmethod
    def _parse_aniboom_response(cls, raw_aniboom_response: str, *, quality: int = 1080, mpd: bool = False) -> str:
        raw_aniboom_response = unescape(raw_aniboom_response)
        if mpd and (url := AniboomDefaults.RE_MPD.findall(raw_aniboom_response)):
            return url[0].replace("\\", "")
        if quality not in AniboomDefaults.QUALITY or quality == 1080:
            return AniboomDefaults.RE_M3U8.findall(raw_aniboom_response)[0].replace("\\", "")
        else:
            return cls._set_quality(AniboomDefaults.RE_M3U8.findall(raw_aniboom_response)[0].replace("\\", ""), quality)

    @classmethod
    def parse(cls, aniboom_player_url: str, *,
              quality: int = 1080,
              mpd: bool = False,
              referer: Optional[str] = None) -> str:

        if not cls.is_aniboom(aniboom_player_url):
            raise TypeError(f"{aniboom_player_url} is not aniboom")
        if not referer:
            referer = AniboomDefaults.REFERER
        resp = cls()._get_aniboom_html_response(aniboom_player_url, referer)
        return cls._parse_aniboom_response(resp, quality=quality, mpd=mpd)

    def __call__(self,
                 aniboom_player_url: str, *,
                 quality: int = 1080,
                 mpd: bool = False,
                 referer: Optional[str] = None) -> str:

        return self.parse(aniboom_player_url, quality=quality, mpd=mpd, referer=referer)



from typing import Any

from anicli_ru.base import *


class Anime(BaseAnimeHTTP):
    """API method write in snake_case style.
    For details see docs on https://github.com/anilibria/docs/blob/master/api_v2.md"""

    BASE_URL = "https://api.anilibria.tv/v2/"

    INSTANT_KEY_REPARSE = True
    _TESTS = {
        "search": ("Зомбиленд", 12),
        "ongoing": True,
        "search_blocked": False,
        "video": True,
        "search_not_found": "_thisTitleIsNotExist123456",
        "instant": "Зомбиленд"
    }

    def api_request(self, *, api_method: str, request_method: str = "GET", **kwargs) -> dict:
        """
        :param str api_method: Anilibria api method
        :param str request_method: requests send method type. Default "POST"
        :param kwargs: any requests.Session kwargs
        :return: json response
        """
        resp = self.session.request(request_method, self.BASE_URL + api_method, **kwargs)
        return resp.json()

    @staticmethod
    def _kwargs_pop_params(kwargs, **params) -> dict:
        data = kwargs.pop("params") if kwargs.get("params") else {}
        data.update(params)
        return data

    def search_titles(self, *, search: str, limit: int = -1, **kwargs) -> dict:
        """searchTitles method

        :param search:
        :param limit:
        :param kwargs:
        :return:
        """
        params = self._kwargs_pop_params(kwargs, search=search, limit=limit)
        return self.api_request(api_method="searchTitles", params=params, **kwargs)

    def get_updates(self, *, limit: int = -1, **kwargs) -> dict:
        """getUpdates method

        :param limit:
        :param kwargs:
        :return:
        """
        params = self._kwargs_pop_params(kwargs, limit=limit)
        return self.api_request(api_method="getUpdates", data=params, **kwargs)

    def episode_reparse(self, *args, **kwargs):
        raise NotImplementedError

    def search(self, q: str) -> ResultList[BaseAnimeResult]:
        return AnimeResult.parse(self.search_titles(search=q))

    def ongoing(self, *args, **kwargs) -> ResultList[BaseOngoing]:
        return Ongoing.parse(self.get_updates())

    def episodes(self, *args, **kwargs) -> ResultList[BaseEpisode]:
        raise NotImplementedError("Get this object from Ongoing or AnimeResult object")

    def players(self, *args, **kwargs) -> ResultList[BasePlayer]:
        raise NotImplementedError("Get this object from Episode object")

    def get_video(self, player_url: str, quality: int = 720, *, referer: str = ""):
        raise NotImplementedError("Get video from Player object")


class Player(BaseJsonParser):
    KEYS = ('key', 'url')
    key: str
    url: str

    def __str__(self):
        return self.key

    def get_video(self, *args, **kwargs) -> str:
        return self.url


class Episode(BaseJsonParser):
    KEYS = ('serie', 'created_timestamp', 'preview', 'skips', 'hls', 'host')
    host: str  # not used in real API response
    serie: int
    created_timestamp: int
    preview: Any
    skips: dict
    hls: dict

    def __str__(self):
        return f"Episode {self.serie}"

    def player(self) -> ResultList[Player]:
        # {url: str, key: str}
        rez = []
        for k, v in self.hls.items():
            if v:  # value maybe equal None
                url = self.host + v if self.host.startswith("http") else f"https://{self.host}{v}"
                rez.extend(Player.parse({"url": url, "key": k}))
        return rez


class AnimeResult(BaseJsonParser):
    KEYS = ('id', 'code', 'names', 'status', 'announce', 'posters', 'updated', 'last_change', 'type',
            'genres', 'team', 'season', 'description', 'in_favorites', 'blocked', 'player', 'torrents')
    id: int
    code: str
    names: dict
    announce: Any
    status: dict
    posters: dict
    updated: int
    last_change: int
    type: dict
    genres: list
    team: dict
    season: dict
    description: str
    in_favorites: int
    blocked: dict
    player: dict
    torrents: dict

    def __str__(self):
        return self.names['ru']

    def episodes(self) -> ResultList[Episode]:
        host = self.player["host"]
        for p in self.player["playlist"].values():
            p["host"] = host
        playlist = list(self.player["playlist"].values())
        return Episode.parse(playlist)


class Ongoing(AnimeResult):
    # response equal AnimeResult object
    pass



"""simple useragent generator"""

from random import choice

CHROME_VERSIONS = (
    '98.0.4758.108', '98.0.4758.107', '98.0.4758.106', '97.0.4692.108', '98.0.4758.105', '98.0.4758.104',
    '98.0.4758.103', '98.0.4758.102', '98.0.4758.101', '98.0.4758.100', '98.0.4758.99', '97.0.4692.107', '98.0.4758.98',
    '96.0.4664.194',
    '98.0.4758.97', '96.0.4664.193', '98.0.4758.96', '96.0.4664.192', '97.0.4692.106', '97.0.4692.105', '96.0.4664.183',
    '97.0.4692.104', '97.0.4692.103', '96.0.4664.181', '96.0.4664.180', '96.0.4664.179', '96.0.4664.178',
    '96.0.4664.177',
    '97.0.4692.102', '96.0.4664.176', '97.0.4692.101', '96.0.4664.175', '97.0.4692.100', '96.0.4664.174',
    '97.0.4692.99',
    '97.0.4692.98', '97.0.4692.97'
)

MOBILE_STRINGS = (
    "(Linux; Android 6.0; Nexus 5)",
    "(Linux; Android 7.0; Redmi Note 7 Pro)",
    "(Linux; Android 8.1.0; Redmi Note 8 Pro)",
    "(Linux; Android 9.0.0; Redmi Note 9 Pro)",
    "(Linux; Android 6.0; SM-A710F)",
    "(Linux; Android 6.0; SAMSUNG SM-C9000)"
)

DESKTOP_STRINGS = (
    "(Windows NT 10.0; Win64; x64)",
    "(Windows NT 11.0; Win64; x64)",
    "(X11; Linux x86_64)",
)


class Agent:
    """Simple UserAgent string generator noname chromium based browser for desktop or mobile

    Basic Usage::

      >>> from anicli_ru.utils.random_agent import Agent
      >>> agent = Agent.desktop()
      Mozilla/5.0 (X11; Linux x86_64) ...

      >>> agent = Agent.mobile()
      Mozilla/5.0 (Linux; Android 6.0; Nexus 5) ...
    """
    @classmethod
    def mobile(cls) -> str:
        """Generate chromium based useragent for mobile

        :return: useragent string
        """
        device = choice(MOBILE_STRINGS)
        chrome = choice(CHROME_VERSIONS)
        return f"Mozilla/5.0 {device} AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{chrome} Mobile Safari/537.36"

    @classmethod
    def desktop(cls) -> str:
        """Generate chromium based useragent for desktop

        :return: useragent string"""
        device = choice(DESKTOP_STRINGS)
        chrome = choice(CHROME_VERSIONS)
        return f"Mozilla/5.0 {device} AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{chrome} Safari/537.36"

    @classmethod
    def random(cls) -> str:
        """Generate desktop or mobile useragent

        :return: useragent string
        """
        return choice((cls.mobile, cls.desktop))()



from __future__ import annotations
from typing import Union
import json

from anicli_ru.base import *


class Anime(BaseAnimeHTTP):
    BASE_URL = "https://api.animevost.org/v1/"

    INSTANT_KEY_REPARSE = True

    def api_request(self, *, api_method: str, request_method: str = "GET", **kwargs) -> dict:
        """
        :param str api_method: Animevost api method
        :param str request_method: requests send method type. Default "POST"
        :param kwargs: any requests.Session kwargs
        :return: json response
        """
        resp = self.session.request(request_method, self.BASE_URL + api_method, **kwargs)
        return resp.json()

    @staticmethod
    def _kwargs_pop_params(kwargs, **params) -> dict:
        data = kwargs.pop("params") if kwargs.get("params") else {}
        data.update(params)
        return data

    def search_titles(self, search: str, **kwargs) -> dict:
        params = self._kwargs_pop_params(kwargs, name=search)
        return self.api_request(api_method="search", request_method="POST", data=params, **kwargs)['data']

    def get_updates(self, *, limit: int = 20, **kwargs) -> dict:
        params = self._kwargs_pop_params(kwargs, page=1, quantity=limit)
        return self.api_request(api_method="last", params=params, **kwargs)['data']

    def episode_reparse(self, *args, **kwargs):
        raise NotImplementedError

    def search(self, q: str) -> ResultList[BaseAnimeResult]:
        return AnimeResult.parse(self.search_titles(search=q))

    def ongoing(self, *args, **kwargs) -> ResultList[BaseOngoing]:
        return Ongoing.parse(self.get_updates())

    def episodes(self, result: Union[AnimeResult, Ongoing], *args, **kwargs) -> ResultList[BaseEpisode]:  # type: ignore
        req = self.api_request(api_method="playlist", request_method="POST", data={'id': result.id})
        return Episode.parse({"episodes": req, "series": result.series})  # signature fix issue

    def players(self, *args, **kwargs) -> ResultList[BasePlayer]:
        raise NotImplementedError("Get this object from Episode object")

    def get_video(self, player_url: str, quality: int = 720, *, referer: str = ""):
        raise NotImplementedError("Get video from Player object")


class Player(BaseJsonParser):
    KEYS = ('key', 'url')
    key: str
    url: str

    def __str__(self):
        return self.key

    def get_video(self, *args, **kwargs) -> str:
        return self.url


class Episode(BaseJsonParser):
    KEYS = ('std', 'preview', 'name', 'hd')
    std: str
    preview: str
    name: str
    hd: str

    @staticmethod
    def sorting_series(series, sort_series):
        name = series.name
        replace = sort_series.replace("\'", "\"")
        jsn = json.loads(replace)
        sort = list(jsn.keys())
        return sort.index(name)

    def __str__(self):
        return self.name

    @classmethod
    def parse(cls, response) -> ResultList:
        """class object factory

        :param response: json response
        :return: ResultList with objects
        """
        # response = sorted(response, key=lambda k: cls.sorting_series(k['name']))
        rez = []
        if isinstance(response["episodes"], list):  # type: ignore
            for data in response["episodes"]:  # type: ignore
                c = cls()
                for k in data.keys():
                    if k in cls.KEYS:
                        setattr(c, k, data[k])
                rez.append(c)
            rez.sort(key=lambda name: cls.sorting_series(name, response["series"]))
        return rez

    def player(self) -> ResultList[Player]:
        rez = []
        rez.extend(Player.parse([{'key': 'hd (720p)', 'url': self.hd},
                                 {'key': 'std (480p)', 'url': self.std}]))
        return rez


class AnimeResult(BaseJsonParser):
    ANIME_HTTP = Anime()
    KEYS = ('id', 'description', 'isFavorite', 'rating', 'series', 'director', 'urlImagePreview', 'year',
            'genre', 'votes', 'title', 'timer', 'type', 'isLikes', 'screenImage')
    id: int
    description: str
    isFavorite: int
    rating: int
    series: dict
    director: str
    urlImagePreview: str
    year: str
    genre: str
    votes: int
    title: str
    timer: int
    type: str
    isLikes: int
    screenImage: list

    def __str__(self):
        return self.title

    def episodes(self):
        with self.ANIME_HTTP as a:
            return a.episodes(self)


class Ongoing(AnimeResult):
    # response equal AnimeResult object
    pass



"""Шаблон для добавления своих парсеров под сторонние сайты

"""
from anicli_ru.base import *
import re


class Anime(BaseAnimeHTTP):
    BASE_URL = "https://example.com"

    def episode_reparse(self, *args, **kwargs):
        # use this method if episodes come first,
        # and then the choice of dubs for the correct operation of the TUI app
        raise NotImplementedError

    def search(self, q: str) -> ResultList[BaseAnimeResult]:
        # Entrypoint for searching anime title by string query

        r = self.session.get(self.BASE_URL).text
        return AnimeResult.parse(r)

    def ongoing(self, *args, **kwargs) -> ResultList[BaseOngoing]:
        # Entrypoint for searching ongoings"""
        r = self.session.get(self.BASE_URL).text
        return Ongoing.parse(r)

    def episodes(self, *args, **kwargs) -> ResultList[BaseEpisode]:
        # for get episodes

        # If the source does not need to send HTTP request, do not override this,
        # but call from the object Ongoing|Episode
        r = self.session.get(self.BASE_URL).text
        return Episode.parse(r)

    def players(self, *args, **kwargs) -> ResultList[BasePlayer]:
        # Entryponint for get video player url
        # If the source does not need to send HTTP request, do not override this, but call from the object Episode

        r = self.session.get(self.BASE_URL).text
        return Player.parse(r)


class AnimeResult(BaseAnimeResult):
    ANIME_HTTP = Anime()
    REGEX = {"url": re.compile("foo (.*?)"),
             "title": re.compile("bar (.*?)")}

    url: str
    title: str

    def __str__(self):
        # return output in terminal or str() func
        return f"{self.title}"


class Ongoing(BaseOngoing):
    ANIME_HTTP = Anime()

    def __str__(self):
        return


class Player(BasePlayer):
    ANIME_HTTP = Anime()
    REGEX = {"url": re.compile("url (.*?)")}
    url: str

    def __str__(self):
        return


class Episode(BaseEpisode):
    ANIME_HTTP = Anime()

    def __str__(self):
        return ""



"""Kodik module utils"""
import re
import warnings
from base64 import b64decode
from typing import Optional, Pattern, Tuple, Dict, List
from urllib.parse import urlparse

from anicli_ru._http import client
from anicli_ru.defaults import KodikDefaults


class Kodik:
    """Class for parse video from kodik balancer

        Example::
            >>> from anicli_ru.kodik import Kodik
            >>> video = Kodik.parse("https://kodik.info/seria/123/hashfoobar123/720p", referer="kodik.info")

        Or with init this class::
            >>> from anicli_ru.kodik import Kodik
            >>> k = Kodik()
            >>> k.parse("https://kodik.info/seria/123/hashfoobar123/720p")
            >>> # or
            >>> k("https://kodik.info/seria/123/hashfoobar123/720p")
    """
    # kodik/anivod etc regular expressions
    DATA_RE = KodikDefaults.data
    KODIK_URL_VALIDATE = KodikDefaults.KODIK_URL_VALIDATE
    KODIK_REFERER = KodikDefaults.RE_KODIK_REFERER
    QUALITY = KodikDefaults.QUALITY

    def __init__(self):
        self.session = client
        self.useragent = self.session.headers.get("user-agent")

    def get_video_url(self, player_url: str, quality: int = 720, *, referer: str = ""):
        warnings.warn("Usage Kodik.parse() or Kodik()(...) methods", category=DeprecationWarning, stacklevel=2)
        return self(player_url, quality, referer=referer)

    @staticmethod
    def decode(url_encoded: str) -> str:
        """kodik player video url decoder (reversed base64 string)

        :param str url_encoded: encoded url
        :return: decoded video url"""
        url_encoded = url_encoded[::-1]
        if not url_encoded.endswith("=="):
            url_encoded += "=="
        link = b64decode(url_encoded).decode()
        if not link.startswith("https"):
            link = f"https:{link}"
        return link

    @staticmethod
    def decode_2(url_encoded: str) -> str:
        # 30.03.23
        # This code replaces all alphabetical characters in a base64 encoded string
        # with characters that are shifted 13 places in the alphabetical order,
        # wrapping around to the beginning or end of the alphabet as necessary.
        # This is a basic form of a Caesar cipher, a type of substitution cipher.

        # Note: this solution write by chatgpt, I don't know how this work :D
        def char_wrapper(e):
            return chr((ord(e.group(0)) + 13 - (65 if e.group(0) <= "Z" else 97))
                       % 26 + (65 if e.group(0) <= "Z" else 97))

        base64_url = re.sub(r"[a-zA-Z]", char_wrapper, url_encoded)
        if not base64_url.endswith("=="):
            base64_url += "=="
        return f"https:{b64decode(base64_url).decode()}"

    @staticmethod
    def is_kodik(url: str) -> bool:
        """return True if url player is kodik"""
        return bool(Kodik.KODIK_URL_VALIDATE.search(url))

    @classmethod
    def parse(cls,
              kodik_player_url: str,
              quality: int = 720,
              *,
              referer: Optional[str] = None,
              raw_response: Optional[str] = None) -> str:
        """
        Class method for get kodik video

        :param kodik_player_url: start kodik player url
        :param quality: video quality. Default None
        :param referer: base source referer. Default None
        :param raw_response: raw response from kodik_player_url object. Default none
        :return: direct video url
        """
        if not cls.is_kodik(kodik_player_url):
            raise TypeError(
                f"Unknown player balancer. get_video_url method support kodik balancer\nvideo url: {kodik_player_url}")
        if not referer:
            referer = f"https://{urlparse(kodik_player_url).netloc}/"
        if not raw_response:
            raw_response = cls()._get_raw_payload(kodik_player_url, referer)

        data, new_referer = cls._parse_payload(raw_response)
        api_url = cls._get_api_url(kodik_player_url)
        video_url = cls()._get_kodik_video_links(api_url, new_referer, data)["360"][0]["src"]  # type: ignore
        return cls()._get_video_quality(video_url, quality)

    def __call__(self, player_url: str, quality: int = 720, *, referer: str = "") -> str:
        """Call method for get video url

        :param str player_url: start_kodik url
        :param int quality: video quality. Default 720
        :param str referer: referer headers. default set auto
        :return: direct video url
        """
        return self.parse(player_url, quality=quality, referer=referer)

    def _get_raw_payload(self, player_url: str, referer: str) -> str:
        return self.session.get(player_url, headers={"user-agent": self.useragent, "referer": referer}).text

    @staticmethod
    def _parse_payload(resp: str) -> Tuple[Dict, str]:
        # sourcery skip: dict-assign-update-to-union

        # prepare values for next POST request
        data = {k: v.findall(resp)[0] if isinstance(v, Pattern) else v for k, v in Kodik.DATA_RE.items()}
        url_data = Kodik.KODIK_REFERER.findall(resp)[0]
        return data, url_data

    @staticmethod
    def _get_api_url(player_url: str):
        if not player_url.startswith("//"):
            player_url = f"//{player_url}"
        if not player_url.startswith("https:"):
            player_url = f"https:{player_url}"

        url_, = Kodik.KODIK_URL_VALIDATE.findall(player_url)
        return f"https://{urlparse(url_).netloc}/gvi"

    def _get_kodik_video_links(self, api_url: str,
                               new_referer: str,
                               data: dict) -> Dict[Dict, List[Dict]]:
        copy_headers = self.session.headers.copy()
        copy_headers.update()

        return self.session.post(api_url, data=data,
                                 headers={"origin": api_url.replace("/gvi", ""), "referer": f"https:{new_referer}",
                                          "accept": "application/json, text/javascript, */*; q=0.01"}).json()["links"]

    def _is_not_404_code(self, url) -> bool:
        return self.session.get(url).status_code != 404

    def _get_video_quality(self, video_url: str, quality: int) -> str:
        quality = 720 if quality not in self.QUALITY else quality
        # video_url = self.decode(video_url)  # outdated
        video_url = self.decode_2(video_url)
        video_url = video_url.replace("360.mp4", f"{quality}.mp4")
        # issue 8, video_url maybe return 404 code
        if self._is_not_404_code(video_url):
            return video_url

        choose_quality = f"{quality}.mp4"

        for q in self.QUALITY:
            video_url = video_url.replace(choose_quality, f"{q}.mp4")
            if self.session.get(video_url).status_code == 200:
                return video_url
            choose_quality = f"{q}.mp4"
        raise RuntimeError("Video not found", video_url)



from __future__ import annotations
from typing import Union, List
from anicli_ru.base import *
import re


class Anime(BaseAnimeHTTP):
    BASE_URL = "https://animania.online/index.php"
    _TESTS = {
        "search": ("experiments lain", 1),
        "ongoing": True,
        "search_blocked": False,
        "video": True,
        "search_not_found": "_thisTitleIsNotExist123456",
        "instant": "experiments lain"
    }

    def search(self, q: str) -> ResultList[BaseAnimeResult]:
        r = self.session.get(self.BASE_URL, params=dict(do="search", subaction="search", story=q))
        return AnimeResult.parse(r.text)

    def ongoing(self) -> ResultList[Ongoing]:  # type: ignore
        r = self.session.get(self.BASE_URL)
        return Ongoing.parse(r.text)

    def episodes(self, result: Union[AnimeResult, Ongoing]) -> ResultList[Episode]:  # type: ignore
        r = self.session.get(result.url, headers=self.session.headers.copy().update(
            {"Referer": result.url}))
        return Episode.parse(r.text)

    def players(self):
        # get players from episode object
        raise NotImplementedError


class AnimeResult(BaseAnimeResult):
    ANIME_HTTP = Anime()
    REGEX = {"url": re.compile(r'<a class="short-poster img-box" href="(.*?\.html)" data-title=".*?: .*?"'),
             "title": re.compile(r'<a class="short-poster img-box" href=".*?\.html" data-title=".*?: (.*?)"')}
    url: str
    title: str

    def __str__(self):
        return f"{self.title}"


class Ongoing(BaseOngoing):
    """
    title: str title name

    num: str episode number

    url: str url
    """
    ANIME_HTTP = Anime()
    REGEX = {
        "raw_url": re.compile(r'<a class="ksupdate_block_list_link" href="(.*?)">.*?</a>'),
        "title": re.compile(r'<a class="ksupdate_block_list_link" href=".*?">(.*?)</a>'),
        "num":
            re.compile(r'<span class="cell cell-2"><a href=".*?">.*?<br>(\d+) .*? </a></span>'),
    }

    title: str
    num: str
    raw_url: str

    @property
    def url(self):
        return f"https://animania.online{self.raw_url}"

    def __str__(self):
        return f"{self.title} {self.num}"


class Player(BasePlayer):
    """
    dub_id int: dubbing ing

    dub_name str: dubbing name

    player str: video player url
    """
    ANIME_HTTP = Anime()
    dub_id: int
    num: int

    @classmethod
    def parse(cls, html: str) -> ResultList:
        raise NotImplementedError("Get <Player> Object from Episode. Ex: Episode().player()")

    def __str__(self):
        u = self._player.replace("//", "").split(".", 1)[0]
        return f"{self.num} {self.dub_name} ({u})"


class Episode(BaseEpisode):
    """
    num: int episode number

    name: str episode name

    id: int episode video id
    """
    ANIME_HTTP = Anime()
    REGEX = {
        "video_chunks": re.compile(
            r"""(<li id="season\d+" style="display:none;">(<span onclick="kodikSlider\.player\('.*?', this\);"> .*?</span>){1,})"""),
        "dubs": re.compile(r"""onclick="kodikSlider\.season\('(\d+)', this\)" style="display:none;">(.*?)</span>"""),
        "videos": re.compile(r"""<span onclick="kodikSlider\.player\('(.*?)', this\);">"""),
        "num": re.compile(r'data-episode="(\d+)"'),
        "name": re.compile(r'data-episode-title="(.*)"'),
    }
    dub_id: int
    dub_name: str
    count: int
    videos: List[str]

    def __str__(self):
        return f"{self.dub_name} count: {self.count}"

    def player(self) -> ResultList[Player]:  # type: ignore
        _players = []
        for i, videos in enumerate(self.videos, 1):
            p: Player = Player()
            p.dub_id, p.dub_name, p._player, p.num = self.dub_id, self.dub_name, videos, i
            _players.append(p)
        return _players

    @classmethod
    def parse(cls, html: str) -> ResultList:
        videos_chunks = cls.REGEX["video_chunks"].findall(html)
        dubs = cls.REGEX["dubs"].findall(html)
        videos = [cls.REGEX["videos"].findall(chunk[0]) for chunk in videos_chunks]
        return [cls(**{"dub_id": dub_id, "dub_name": dub_name, "count": count, "videos": video})
                for dub_id, dub_name, count, video in zip([int(n[0]) for n in dubs],
                                                          [n[1] for n in dubs],
                                                          [len(n) for n in videos], videos)]  # dub id  # dub name  # count videos

    def __eq__(self, other):
        return self.count == other.count

    def __ge__(self, other):
        return self.count >= other.count

    def __le__(self, other):
        return self.count > other.count



"""configured session object with ddos protect detector"""
import warnings

from requests import Session, Response

from anicli_ru.defaults import DDOS_SERVICES, BASE_HEADERS_DICT

__all__ = ('client', 'SessionM')


class SessionM(Session):
    # add custom timeout param https://github.com/psf/requests/issues/3070
    # https://stackoverflow.com/a/62044757
    def __init__(self, timeout=60):
        self.timeout = timeout
        super().__init__()

    def request(self, method, url, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = self.timeout
        return super().request(method, url, **kwargs)


def check_ddos_protect_hook(resp: Response, **kwargs):
    if resp.headers.get("Server") in DDOS_SERVICES \
            and resp.headers["Connection"] == 'close' \
            or resp.status_code == 403:

        warnings.warn(f"{resp.url} have ddos protect and return 403 code.",
                      category=RuntimeWarning,
                      stacklevel=2)


client = SessionM()
client.headers.update(BASE_HEADERS_DICT)
client.hooks["response"] = [check_ddos_protect_hook]



from __future__ import annotations

import warnings
import re
from html import unescape
from typing import Optional, Dict, Pattern, Sequence, Union, List, TypedDict, Tuple

from requests import Response

from anicli_ru._http import client
from anicli_ru import Kodik, Aniboom

__all__ = ("BaseAnimeHTTP",
           "BaseParser",
           "BaseJsonParser",
           "BaseAnimeResult",
           "BasePlayer",
           "BaseEpisode",
           "BaseOngoing",
           "ResultList")


class TESTS(TypedDict):
    """ Dict for testing extractors

    search: title_name: str, episodes: int - check search method and get episodes

    ongoing: bool - check ongoings search

    video: bool - test get video url

    search_blocked: bool - ignore failed get episode and retry get episodes for non blocked title

    search_not_found: str title - this title has not exist

    instant: str - title, where need get all video urls

    """
    search: Tuple[str, int]
    ongoing: bool
    video: bool
    search_blocked: bool
    search_not_found: str
    instant: str


class BaseAnimeHTTP:
    """Base singleton class for send request to url, where need get html/json documents.

    the following variables and methods should be overridden in this class:

    BASE_URL: - main url site

    def search(self, q: str): - search by string

    def ongoing(self, *args, **kwargs): - search ongoings

    def episodes(self, *args, **kwargs): - search episodes

    def players(self, *args, **kwargs): - search direct video url

    _TESTS: - config dict for testing extractor

    Optional:

        USER_AGENT: - Useragent. In CLI, default set random by utils.random_agent module


    """
    BASE_URL = "https://example.com/"
    # XMLHttpRequest value required!
    USER_AGENT = {
        "user-agent":
            "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) "
            "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.114 Mobile Safari/537.36",
            "x-requested-with": "XMLHttpRequest"}
    _instance = None
    TIMEOUT: float = 30
    # dict for get parser config tests
    _TESTS: TESTS = {
        "search": ("experiments lain", 13),  # standard search test
        "ongoing": True,  # test search ongoings, True - yes, False - no
        "video": True,  # test get raw video, True - yes, False - no
        "search_blocked": False,  # ignore failed get episode and retry get episodes for non blocked title
        "search_not_found": "_thisTitleIsNotExist123456",  # this title has not exist
        "instant": "experiments lain"  # test instant key scroll series
    }
    # костыль для настройки поведения ключа INSTANT issue #6:
    # если сначала идёт выбор озвучки, а потом плеера, выставите значение True (see extractors/animego)
    INSTANT_KEY_REPARSE = False

    def __new__(cls, *args, **kwargs):
        # create singleton for correct store session
        if not cls._instance:
            cls._instance = super(BaseAnimeHTTP, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        self.session = client
        self.session.timeout = self.TIMEOUT
        self.session.headers.update(self.USER_AGENT)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()

    # def request(self, method: str, url: str, **kwargs) -> Response:
    #     """Session.request method
    #
    #     :param str method: method type
    #     :param str url: url target
    #     :param kwargs: optional requests.Session kwargs
    #     :return: requests.Response object
    #     """
    #     # context manager solve ResourceWarning (trace this in tests)
    #     warnings.warn("Usage self.session.request method in parsers", category=SyntaxWarning, stacklevel=2)
    #     with self.session as s:
    #         return s.request(method, url, timeout=self.TIMEOUT, **kwargs)
    #
    # def request_get(self, url: str, **kwargs) -> Response:
    #     """Session.get method
    #
    #     :param str url: url target
    #     :param kwargs: optional requests.Session kwargs
    #     :return: requests.Response object
    #     """
    #     warnings.warn("Usage self.session.get method in parsers", category=SyntaxWarning, stacklevel=2)
    #     return self.request("GET", url, **kwargs)
    #
    # def request_post(self, url: str, **kwargs) -> Response:
    #     """Session.post method
    #
    #     :param url: url target
    #     :param kwargs: optional requests.Session kwargs
    #     :return: requests.Response object
    #     """
    #     warnings.warn("Usage self.session.post method in parsers", category=SyntaxWarning, stacklevel=2)
    #     return self.request("POST", url, **kwargs)

    # need manually write requests in parsers with self.session object

    def search(self, q: str) -> ResultList[BaseAnimeResult]:
        """Search anime title by string pattern

        :param str q: string search
        :return: list with AnimeResult objects
        """
        raise NotImplementedError

    def ongoing(self, *args, **kwargs) -> ResultList[BaseOngoing]:
        """Search ongoings

        :param args:
        :param kwargs:
        :return: list with Ongoing objects
        """
        raise NotImplementedError

    def episodes(self, *args, **kwargs) -> ResultList[BaseEpisode]:  # type: ignore
        """

        :param args:
        :param kwargs:
        :return: list with Episode objects
        """
        raise NotImplementedError

    def episode_reparse(self, *args, **kwargs):
        """Need write this method if INSTANT_KEY_REPARSE == True (see, extractors/animego.py)

        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def players(self, *args, **kwargs) -> ResultList[BasePlayer]:  # type: ignore
        """

        :param args:
        :param kwargs:
        :return: list with Player objects
        """
        raise NotImplementedError

    def get_kodik_video(self, player_url: str, quality: int = 720, *, referer: str = "") -> str:
        """Get hls url from kodik balancer

        :param str player_url: - raw url from kodik balancer
        :param int quality: - video quality. Default 720
        :param str referer: - referer, where give this url. Default get automatically
        :return: direct video url
        """
        return Kodik()(player_url, quality, referer=referer)

    def get_aniboom_video(self, player_url: str, quality: int = 1080) -> str:
        """get hls url from aniboom balancer

        :param player_url: raw url from aniboom balancer
        :return: direct video url
        """
        # fix 28 11 2021 request
        referer = self.BASE_URL if self.BASE_URL.endswith("/") else f"{self.BASE_URL}/"
        return Aniboom().get_video_url(player_url, referer=referer, quality=quality)

    def get_video(self, player_url: str, quality: int = 720, *, referer: str = "") -> str:
        """Get video from balancer. Check balancer, where from url

        :param player_url: player url from any balancer
        :param quality: video quality. Default 720
        :param referer: referer string. Default set automatically
        :return:
        """
        if "sibnet" in player_url:
            return player_url
        elif Aniboom.is_aniboom(player_url):
            url = self.get_aniboom_video(player_url, quality=quality)
            return url
        elif Kodik.is_kodik(player_url):
            url = self.get_kodik_video(player_url, quality, referer=referer)
            return url
        else:
            # catch any players for add in script
            raise TypeError(f"The link {player_url} is not defined by the available balancers.")


# old alias
ResultList = List


class BaseParser:
    """Base object parser from text response

    REGEX: Dict[key: str, value: Pattern] - a dictionary with keys, and
    re.compile expressions values, which the parser will get and assign to objects

    re.compile expression must extract **one group value** from the document for the default parser to work correctly.

    If this is not the case, then you must redefine the logic of the parse classmethod
    """

    REGEX: Dict[str, Pattern]  # {"attr_name": re.compile(<regular expression>)}

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, str):
                v = int(v) if v.isdigit() else unescape(str(v))
            setattr(self, k, v)

    @classmethod
    def parse(cls, html: str) -> ResultList:
        """class object factory

        :param str html: html document
        :return: ResultList with objects
        """
        l_objects = []
        # generate dict like {attr_name: list(values)}
        results = {k: re.findall(v, html) for k, v in cls.REGEX.items()}
        for values in zip(*results.values()):
            attrs = zip(results.keys(), values)
            # generate objects like {attr_name: attr_value}
            l_objects.append(cls(**dict(attrs)))
        return l_objects


class BaseJsonParser:
    """Base parser object for JSON response (see extractors/anilibria.py)

    KEYS: Sequence - collection of keys to get from json"""
    REGEX = None
    KEYS: Sequence[str]

    @classmethod
    def parse(cls, response: Union[Dict, List]) -> ResultList:
        """class object factory

        :param response: json response
        :return: ResultList with objects
        """
        rez = []
        if isinstance(response, list):
            for data in response:
                c = cls()
                for k in data.keys():
                    if k in cls.KEYS:
                        setattr(c, k, data[k])
                rez.append(c)
        elif isinstance(response, dict):
            c = cls()
            for k in response.keys():
                if k in cls.KEYS:
                    setattr(c, k, response[k])
            rez.append(c)
        return rez


BaseParserObject = BaseParser  # old alias


class BasePlayer(BaseParserObject):
    """Player object class. Contains balancer url and method get video url

    """
    ANIME_HTTP: BaseAnimeHTTP
    dub_name: str
    _player: str

    @property
    def url(self) -> str:
        """get player url"""
        return self.player_prettify(self._player)

    @staticmethod
    def player_prettify(player: str):
        """Player prettify url method and convert all named and numeric character references
        (e.g. &gt;, &#62;, &x3e;) in the string to the corresponding unicode characters:


        //foobar.com/barfoo12345 -> https://foobar.com/barfoo12345


        :param str player: - raw player url
        :return:
        """
        return f"https:{unescape(player)}"

    def get_video(self, quality: int = 720, referer: Optional[str] = None) -> str:
        """Get direct video url

        :param int quality: video quality. Default 720
        :param referer: referer string. Default set auto.
        :return: video url
        """
        if not referer:
            referer = self.ANIME_HTTP.BASE_URL if self.ANIME_HTTP.BASE_URL.endswith("/") else f"{self.ANIME_HTTP.BASE_URL}/"

        with self.ANIME_HTTP as a:
            return a.get_video(player_url=self.url, quality=quality, referer=referer)


class BaseEpisode(BaseParserObject):
    ANIME_HTTP: BaseAnimeHTTP

    def player(self) -> ResultList[BasePlayer]:
        """Get list with Player object

        :return: ResultList with Player objects
        """
        with self.ANIME_HTTP as a:
            return a.players(self)


class BaseOngoing(BaseParserObject):
    ANIME_HTTP: BaseAnimeHTTP
    url: str
    title: str

    def episodes(self) -> ResultList[BaseEpisode]:
        """Get list with Episode objects

        :return: ResultList with Episode objects
        """
        with self.ANIME_HTTP as a:
            return a.episodes(self)


class BaseAnimeResult(BaseParserObject):
    ANIME_HTTP: BaseAnimeHTTP
    url: str
    title: str

    def episodes(self) -> ResultList[BaseEpisode]:
        """Get list with Episode objects

        :return: ResultList with Episode objects
        """
        with self.ANIME_HTTP as a:
            return a.episodes(self)



from __future__ import annotations
from typing import Union, List
from anicli_ru.base import *
import re


class Anime(BaseAnimeHTTP):
    BASE_URL = "https://animego.org/"
    INSTANT_KEY_REPARSE = True

    def search(self, q: str) -> ResultList[AnimeResult]:  # type: ignore
        resp = self.session.get(f"{self.BASE_URL}search/anime", params={"q": q}).text
        return AnimeResult.parse(resp)

    def ongoing(self) -> ResultList[Ongoing]:  # type: ignore
        resp = self.session.get(self.BASE_URL).text
        return Ongoing.parse(resp)

    def episodes(self, result: Union[AnimeResult, Ongoing]) -> ResultList[Episode]:  # type: ignore
        resp = self.session.get(f"{self.BASE_URL}anime/{result.id}/player?_allow=true").json()["content"]
        return Episode.parse(resp)

    def players(self, episode: Episode) -> ResultList[Player]:  # type: ignore
        resp = self.session.get(f"{self.BASE_URL}anime/series",
                                params={"dubbing": 2, "provider": 24,
                                        "episode": episode.num, "id": episode.id}).json()["content"]
        return Player.parse(resp)


class AnimeResult(BaseAnimeResult):
    REGEX = {"url": re.compile(r'<a href="(https://animego\.org/anime/.*)" title=".*?">'),
             "title": re.compile(r'<a href="https://animego\.org/anime/.*" title="(.*?)">')}
    ANIME_HTTP = Anime()
    url: str
    title: str

    @property
    def id(self) -> str:
        return self.url.split("-")[-1]

    def __str__(self):
        return f"{self.title}"


class Ongoing(BaseOngoing):
    ANIME_HTTP = Anime()
    REGEX = {
        "raw_url": re.compile(r'onclick="location\.href=\'(.*?)\'"'),
        "title": re.compile(r'600">(.*?)</span></span></div><div class="ml-3 text-right">'),
        "num":
            re.compile(r'<div class="font-weight-600 text-truncate">(\d+) серия</div><div class="text-gray-dark-6">'),
        "dub": re.compile(r'<div class="text-gray-dark-6">(\(.*?\))</div>'),
        "thumbnail": re.compile(r'"background-image: url\((.*?)\);"></div>')
    }

    title: str
    num: str
    dub: str
    raw_url: str

    thumbnail: str

    @classmethod
    def parse(cls, html: str) -> ResultList:
        ongoings = []
        # generate dict like {attr_name: list(values)}
        results = {k: re.findall(v, html) for k, v in cls.REGEX.items()}

        for values in zip(*results.values()):
            attrs = zip(results.keys(), values)
            # generate objects like {attr_name: attr_value}
            ongoings.append(cls(**dict(attrs)))

        # shitty sort duplicates (by title and episode num) algorithm
        # but ongoings list contains less than 100 elements guaranty
        sorted_ongoings: List[Ongoing] = []
        for ongoing in ongoings:
            if ongoing in sorted_ongoings:
                for sorted_ong in sorted_ongoings:
                    if ongoing == sorted_ong:
                        sorted_ong += ongoing
                        break
            else:
                sorted_ongoings.append(ongoing)

        sorted_ongoings.sort(key=lambda k: k.title)
        return sorted_ongoings

    @property
    def url(self):
        return f"https://animego.org{self.raw_url}"

    @property
    def id(self) -> str:
        return self.url.split("-")[-1]

    def __eq__(self, other):
        """return True if title name and episode num equals"""
        return self.num == other.num and self.title == other.title and other.dub not in self.dub

    def __iadd__(self, other):
        """Add dub name in ongoing object with += operator"""
        self.dub += f" {other.dub}"
        return self

    def __add__(self, other):
        """Add dub name in ongoing object with + operator"""
        self.dub += f" {other.dub}"
        return self

    def __str__(self):
        return f"{self.title} {self.num} {self.dub}"


class Episode(BaseEpisode):
    ANIME_HTTP = Anime()
    REGEX = {"num": re.compile(r'data-episode="(\d+)"'),
             "id": re.compile(r'data-id="(\d+)"'),
             "name": re.compile(r'data-episode-title="(.*)"')
             }
    num: int
    name: str
    id: int

    def __str__(self):
        return f"{self.name}"


class Player(BasePlayer):
    ANIME_HTTP = Anime()
    REGEX = {
        "dub_name":
            re.compile(
                r'data-dubbing="(\d+)"><span class="video-player-toggle-item-name text-underline-hover">\s+(.*)'),
        "player":
            re.compile(
                r'data-player="(.*?)"\s+data-provider="\d+"\s+data-provide-dubbing="(\d+)"'),
        "dub_id": re.compile(r"")
    }
    _all_raw_player_urls: list
    dub_name: str
    _player: str
    dub_id: int

    @classmethod
    def parse(cls, html: str) -> ResultList[Player]:
        l_objects = []
        # generate dict like {attr_name: list(values)}
        dub_names = re.findall(cls.REGEX["dub_name"], html)  # dub_id, dub_name
        players = re.findall(cls.REGEX["player"], html)  # player_url, dub_id
        for player, dub_id_1 in players:
            p = Player()
            for dub_id_2, dub_name in dub_names:
                # removed check for catching unsupported players
                if dub_id_1 == dub_id_2:
                    p._player, p.dub_name, p.dub_id = player, dub_name, dub_id_1
                    l_objects.append(p)
        return l_objects

    @property
    def url(self) -> str:
        return self.player_prettify(self._player)

    def __str__(self):
        u = self._player.replace("//", "").split(".", 1)[0]
        return f"{self.dub_name} ({u})"



import argparse
from os import system
from typing import Optional

from anicli_ru.loader import all_extractors
from anicli_ru.__version__ import __version__
from anicli_ru.utils import Agent

ALL_PARSERS = dict(enumerate(all_extractors()))


def setup_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=f"anicli-ru {__version__}\n"
                                                 f"See detail info: https://github.com/vypivshiy/ani-cli-ru",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("-d", "--download",
                        dest="DOWNLOAD",
                        default=False,
                        action="store_true",
                        help="Download mode. Default False. ffmpeg required")
    parser.add_argument("-q", "--quality",
                        dest="QUALITY",
                        type=int,
                        default=1080,
                        choices=[360, 480, 720, 1080],
                        help="Default video quality. Kodik balancer max 720 quality (usually 480), "
                             "Aniboom max 1080, Sibnet 360")
    parser.add_argument("-i", "--instant",
                        dest="INSTANT",
                        default=False,
                        action="store_true",
                        help="Instant view mode. Useful if you want to watch a title without manual switching episodes")
    parser.add_argument("-s", "--source",
                        dest="SOURCE",
                        type=int,
                        default=0,
                        choices=[*range(len(ALL_PARSERS))],
                        help="Site source keys: {}... DEFAULT 0. "
                             "Usage --print-sources for get available parsers".format(
                            ', '.join((str(i) + ' - ' + str(p) for i, p in ALL_PARSERS.items() if i < 3))))
    parser.add_argument("--print-sources",
                        dest="PRINT_SOURCES",
                        default=False,
                        action="store_true",
                        help="Print available source parsers")

    parser.add_argument("-V", "--version",
                        dest="PRINT_VERSION",
                        help="Display version information.",
                        action="version",
                        version=f"anicli-ru {__version__}")

    request_group = parser.add_argument_group("Configure requests", "Options for config request.Session object")
    request_group.add_argument("--disable-agent",
                               dest="RANDOM_AGENT",
                               action="store_true",
                               default=True,
                               help="Disable set random user agent")
    request_group.add_argument("--agent-type",
                               dest="RANDOM_AGENT_TYPE",
                               type=str,
                               choices=["m", "d", "r", "f"],
                               default="m",
                               help="Set randomize agent type. 'm' - mobile, 'd' - desktop, 'r' - random,"
                                    " 'f' - disable randomize. Default 'm'")
    request_group.add_argument("--user-agent",
                               dest="USERAGENT",
                               type=str,
                               default="",
                               help="Set user-agent string")
    request_group.add_argument("--timeout",
                               dest="TIMEOUT",
                               type=float,
                               default=30,
                               help="Max request timeout value. Default 30 seconds")
    request_group.add_argument("-p", "--proxy",
                               dest="PROXY",
                               type=str,
                               default="",
                               help="Add proxy for search requests (not for download video in video player). "
                                    "Didn't have auto detect proxy type. Write argument like "
                                    "proxy_type://ip:port@login:pass")

    updater_group = parser.add_argument_group("Update script", "Check and update script from pypi")
    updater_group.add_argument("-U", "--upgrade",
                               dest="UPGRADE",
                               default=False,
                               action="store_true",
                               help="Upgrade script from pypi")
    updater_group.add_argument("-F", "--force",
                               dest="FORCE",
                               default=False,
                               action="store_true",
                               help="Force update script")
    args = parser.parse_args()
    check_args(args)
    return args


def print_sources():
    """print available anime parsers"""
    for k, v in ALL_PARSERS.items():
        print(f"[{k}] {v}")
    exit(0)


def get_agent(key: str) -> Optional[str]:
    if key == "m":
        return Agent.mobile()
    elif key == "d":
        return Agent.desktop()
    elif key == "f":
        return None
    return Agent.random()


def get_updates(pypi_package_name: str = "anicli-ru", force: bool = False) -> None:
    """Updater function.
    Download last update for pypi.

    :param str pypi_package_name: pypi package name
    :param bool force: force update. Default False
    """
    import requests
    print("Check updates")
    r = requests.get(f"https://pypi.org/pypi/{pypi_package_name}/json/",
                     headers={"user-agent": f"anicli-ru {__version__} Updater"}).json()
    pypi_version = list(r["releases"].keys())[-1]
    if __version__ != pypi_version:
        print(f"Detect new version (your - {__version__}, pypi - {pypi_version})")
    else:
        print("Used last version")
    if __version__ != pypi_version or force:
        answer = input("Update? (y/n)? ")
        if answer.lower() != "y":
            exit(1)
        print("Start update from pypi")
        system(f"python3 -m pip install {pypi_package_name} -U")
        print("Done.")
    exit(0)


def check_args(args: argparse.Namespace) -> None:
    """check specify optional arguments"""
    if args.UPGRADE:
        get_updates(force=args.FORCE)
    elif args.PRINT_SOURCES:
        print_sources()



"""Default constants for this project"""
import re

# antiddos services strings detect
from typing import NamedTuple

DDOS_SERVICES = ("cloudflare", "ddos-guard")

# default user-agent for all project
USER_AGENT = "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, " \
             "like Gecko) Chrome/94.0.4606.114 Mobile Safari/537.36 "

BASE_HEADERS_DICT = {"user-agent": USER_AGENT,
                     "x-requested-with": "XMLHttpRequest"}


# outdated signatures for kodik
# RE_URL = re.compile(r"https://\w+\.\w{2,6}/seria/\d+/\w+/\d{3,4}p")
# RE_URL_DATA = re.compile(r'iframe.src = "//(.*?)"')
# RE_VIDEO_TYPE = re.compile(r"go/(\w+)/\d+")
# RE_VIDEO_ID = re.compile(r"go/\w+/(\d+)")
# RE_VIDEO_HASH = re.compile(r"go/\w+/\d+/(.*?)/\d+p\?")

# kodik.info http request headers 12.07.22
# :method: POST
# :path: /gvi
# accept: application/json, text/javascript, */*; q=0.01
# accept-encoding: gzip, deflate, br
# content-type: application/x-www-form-urlencoded; charset=UTF-8
# origin: https://kodik.info
# referer: https://kodik.info/seria/1234567/hash123foobar/720p?translations=false
# user-agent: Mozilla/5.0 (X11; Linux x86_64) ...
# x-requested-with: XMLHttpRequest


class KodikDefaults:
    # pattern for parse payload next /gvi entrypoint POST request
    data = dict(
        d=re.compile(r'"d":"(.*?)"'),
        d_sign=re.compile(r'"pd":"(.*?)"'),
        pd_sign=re.compile(r'"pd_sign":"(.*?)"'),
        ref=re.compile(r'"ref":"(.*?)"'),
        ref_sign=re.compile(r'"ref_sign":"(.*?)"'),
        hash=re.compile(r"videoInfo\.hash = '(\w+)';"),
        id=re.compile(r'var videoId = "(\d+)"'),
        type=re.compile(r'var type = "(\w+)";'),
        bad_user=True,
        info="{}"
    )
    QUALITY = (720, 480, 360)
    RE_KODIK_REFERER = re.compile(r"data-code='(//\w+.\w{1,6}/\w{3,15}/\d+/\w+/\d{3}p)'>")
    KODIK_URL_VALIDATE = re.compile(r"//\w+\.\w{1,8}/\w{2,24}/\d+/\w+/")
    # default referer
    REFERER = "https://kodik.info/"


# aniboom.one request headers 12.07.22
# :authority: aniboom.one
# :method: GET
# :path: /embed/hash123foobar?episode=1&translation=2
# accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9
# accept-language: ru-RU,ru
# referer: https://animego.org/
# upgrade-insecure-requests: 1
# user-agent: Mozilla/5.0 (X11; Linux x86_64) ...


class AniboomDefaults:
    RE_M3U8 = re.compile(r'"hls":"{\\"src\\":\\"(.*\.m3u8)\\"')
    RE_MPD = re.compile(r'"{\\"src\\":\\"(.*\.mpd)\\"')
    # parse master.m3u8, return [(quality, url), ...] results
    RE_M3U8_DATA = re.compile(
        r'''#EXT-X-STREAM-INF:BANDWIDTH=\d+,RESOLUTION=(\d+x\d+),CODECS=".*?",AUDIO=".*?"\s(.*?\.m3u8)''')
    QUALITY = (1080, 720, 480, 360)  # works only m3u8 format
    REFERER = "https://aniboom.one/"
    ACCEPT_LANG = "ru-RU"


# # Aniboom M3U8 object
class AniboomM3U8Data(NamedTuple):
    quality: str
    url_suffix: str



#!/usr/bin/python3
# mypy: ignore-errors
"""main cli script"""

from __future__ import annotations

import subprocess
from typing import Sequence
from os import name as sys_name
from os import system
from random import sample
from string import ascii_letters
from typing import Union, Optional

from .loader import import_extractor
from .options import ALL_PARSERS, setup_arguments, get_agent
from anicli_ru import Aniboom

args = setup_arguments()

PLAYER = "mpv"
OS_HEADERS_COMMAND = "http-header-fields"

# load chosen extractor
extractor = f"anicli_ru.extractors.{ALL_PARSERS.get(args.SOURCE)}"
API = import_extractor(extractor)

if not args.UPGRADE and not args.FORCE:
    print("Chosen source:", API.Anime.BASE_URL)


class Menu:
    INSTANT = args.INSTANT
    DOWNLOAD = args.DOWNLOAD
    TIMEOUT = args.TIMEOUT

    def __init__(self):
        self.__ACTIONS = {"b": ("[b]ack next step", self.back_on),
                          "c": ("[c]lear console", self.cls),
                          "h": ("[h]elp", self.help),
                          "o": ("[o]ngoing print", self.ongoing),
                          "q": ("[q]uit", self.exit),
                          "e": ("[e]xit (alias quit)", self.exit),
                          }
        self.anime = API.Anime()
        if args.USERAGENT:
            self.anime.session.headers.update({"user-agent": args.USERAGENT})
        elif args.RANDOM_AGENT:
            if agent := get_agent(args.RANDOM_AGENT_TYPE):
                self.anime.session.headers.update({"user-agent": agent})
        self.anime.TIMEOUT = self.TIMEOUT
        self.__back_action = True

    def back_on(self):
        self.__back_action = False

    def back_off(self):
        self.__back_action = True

    @property
    def is_back(self):
        return self.__back_action

    def command_wrapper(self, command) -> bool:
        self.cls()
        if self.__ACTIONS.get(command):
            self.__ACTIONS[command][1]()
            return True
        return False

    def ongoing(self):
        while self.is_back:
            ongoings = self.anime.ongoing()
            if len(ongoings) > 0:
                self._print_enumerate(ongoings)
                print("Choose anime:", 1, "-", len(ongoings))
                command = input(f"c_o [1-{len(ongoings)}] > ")
                if self.command_is_digit(command):
                    self.choose_episode(ongoings[int(command) - 1])
            else:
                print("Cannot grub ongoings, try running this script with the -s argument or use VPN/proxy")
                return
        self.back_off()

    def choose_dub(self, results: API.ResultList[API.Player]):
        while self.is_back:
            self._print_enumerate(results)
            print("Choose dub:", 1, "-", len(results))
            command = input(f"c_d [1-{len(results)}] > ")
            if self.command_is_digit(command):
                command = int(command)
                if command <= len(results):
                    print("Start playing")
                    if self.INSTANT:
                        self._run_instant(start=command, players=results)
                    else:
                        self._run_video(results[command - 1])
        self.back_off()

    def choose_episode(self, result: Union[API.AnimeResult, API.Ongoing]):
        episodes = result.episodes()
        if len(episodes) > 0:
            while self.is_back:
                self._print_enumerate(episodes)
                print(f"Choose episode: 1-{len(episodes)}")
                command = input(f"c_e [1-{len(episodes)}] > ")
                if self.command_is_digit(command):
                    command = int(command)
                    if self.anime.INSTANT_KEY_REPARSE and self.INSTANT:
                        self.episode_instant(episodes, command)
                    elif command <= len(episodes):
                        results = episodes[command - 1].player()
                        if len(results) > 0:
                            self.choose_dub(results)
                        else:
                            print("No available dubs")
                            return
            self.back_off()
        else:
            print(
                "Episodes not found :(\nThis anime-title maybe blocked in your country, try using a vpn/proxy or change "
                "source with -s argument and repeat operation")
        return

    def episode_instant(self, episodes: API.ResultList[API.Episode], start: int):
        # КОСТЫЛЬ!!! issue 6: correct run instant play
        players = episodes[start - 1].player()
        while self.is_back:
            self._print_enumerate(players)
            print("Choose dub:", 1, "-", len(players))
            command = input(f"c_d [1-{len(players)}] > ")
            if self.command_is_digit(command):
                command = int(command)
                if command <= len(players):
                    # get player for compare by name (__str__ method)
                    player_name = str(players[command - 1])
                    for episode in episodes[start - 1:]:
                        for player in episode.player():
                            if str(player) == player_name:
                                print(f"Run '{episode}'")
                                self._run_video(player)
                                break
                    break
        self.back_off()

    def choose_anime(self, results: API.ResultList[API.AnimeResult]):
        while self.is_back:
            self._print_enumerate(results)
            print("Choose anime:", 1, "-", len(results))
            command = input(f"c_a [1-{len(results)}] > ")
            if self.command_is_digit(command):
                command = int(command)
                if 0 < command <= len(results):
                    self.choose_episode(results[command - 1])
                return
        self.back_off()

    def find(self, pattern):
        results = self.anime.search(pattern)
        if len(results) > 0:
            print("Found", len(results))
            self.choose_anime(results)
        else:
            print("Not found!")
            return

    def main(self):
        if args.PROXY:
            self.proxy()
        while True:
            print("Input anime name or USAGE: h for get commands")
            command = input("m > ")
            if not self.command_wrapper(command):
                self.find(command)

    def _run_instant(self, start: int, players: API.ResultList[API.Player]):
        for player in players[start - 1:]:
            self._run_video(player)

    def _run_video(self, player: API.Player):
        url = player.get_video(quality=args.QUALITY)
        if self.DOWNLOAD:
            self._run_download(url)

        elif Aniboom.is_aniboom(url):
            # Экспериментально выявлено одним из пользователей,
            # что заголовок Accept-Language увеличивает скорость загрузки в MPV плеере в данном балансере
            run_player(url, commands= (f"-{OS_HEADERS_COMMAND}", "Referer: https://aniboom.one,Accept-Language: ru-RU, User-Agent:{self.anime.USER_AGENT['user-agent']}"))
        else:
            run_player(url)

    def _run_download(self, player_url: str):
        if Aniboom.is_aniboom(player_url):
            self._download(
                f'ffmpeg -y -headers "Referer: https://aniboom.one,Accept-Language: ru-RU",'
                f'User-Agent:{self.anime.USER_AGENT["user-agent"]}"'
                f'-i "{player_url}" "{"".join(sample(ascii_letters, 12))}.mp4"')
        else:
            self._download(f'ffmpeg -y -i "{player_url}" "{"".join(sample(ascii_letters, 12))}.mp4"')

    @staticmethod
    def cls():
        system("cls") if sys_name == 'nt' else system("clear")

    @staticmethod
    def _download(command: str):
        system(command)

    @staticmethod
    def _print_enumerate(itr: Sequence, *args):
        """print elements with getattr names arg. Default invoke __str__ method"""
        if len(itr) > 0:
            for i, obj in enumerate(itr, 1):
                if args:
                    print(f"[{i}]", *(getattr(obj, arg) for arg in args))
                else:
                    print(f"[{i}]", obj)
        else:
            print("Results not found!")

    def proxy(self):
        print("Check proxy")
        try:
            self.anime.session.get(self.anime.BASE_URL, proxies=dict(http=args.PROXY, https=args.PROXY),
                                   timeout=10)
        except Exception as e:
            print(e.__class__.__name__, "Failed proxy connect")
            self.exit()
        self.anime.session.proxies.update(dict(http=args.PROXY, https=args.PROXY))
        print("Proxy connect success")

    def command_is_digit(self, command: str) -> bool:
        return not self.command_wrapper(command) and command.isdigit()

    @staticmethod
    def exit():
        print("Invoke exit command")
        exit(1)

    def help(self):
        for k, v in self.__ACTIONS.items():
            print(k, v[0])

    @classmethod
    def run(cls):
        cls().main()

    @classmethod
    def validate(cls) -> None:
        try:
            cls().anime.session.get(cls().anime.BASE_URL)
        except OSError as e:
            raise ConnectionError("Connection aborted (Reset by peer). Use Another extractor. "
                                  "To see all available extractors id usage: anicli-ru --print-sources") from e


def run_player(url: str, player: str = None, commands: tuple[Optional[str], ...] = ()) -> None:
    if not player:
        player = PLAYER
    if commands:
        subprocess.run([player, url, *commands])  # type: ignore
    else:
        subprocess.run([player, url])


def main():
    # check status code:
    Menu.validate()
    try:
        Menu.run()
    except KeyboardInterrupt:
        print("KeyboardInterrupt, Exit...")
        exit(1)


if __name__ == '__main__':
    main()
